Object subclass: #EnchantingCodeExporter
	instanceVariableNames: 'dict functions indentationLevel additionalIndent sprite codeStream argDepth template functionIndex outputFileBaseName '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Enchanting'!
!EnchantingCodeExporter commentStamp: 'CWB 8/21/2010 10:15' prior: 0!
The EnchantingCodeExported is used to take the scripts of a Scratch sprite and export them into computer code.

dict is a dictionary to convert from the English representation of a code tile to whichever language interests us.

functions is a list of enchanting functions to be exported.

indentationLevel is how many tabs to put in front of a line of code.
additionalIndent is indentation read from a template file, and is prepended to the indentation.

sprite points back to the sprite whose code tiles we are exporting.

codeStream is a stream for placing the code sequences we are building on.

When argDepth is greater than zero, we are within an argument block.  (No tabbing should be performed).

template is the lines of the template file.

functionIndex assigns a number to each function/thread.


!


!EnchantingCodeExporter methodsFor: 'initialization' stamp: 'CWB 8/21/2010 10:15'!
initialize

	super initialize.
	functions _ Dictionary new.
	indentationLevel _ 0.
	additionalIndent _ ''.
	argDepth _ 0.
	sprite _ nil.
	functionIndex _ 1.! !

!EnchantingCodeExporter methodsFor: 'initialization' stamp: 'CWB 8/26/2010 20:02'!
outputFileBaseName: aName
	"Sets the base file name.  If the full name if /path/to/foo.script, the base name is foo."
	"This is needed as in Java, the file is named after the class"

	outputFileBaseName _ aName.! !

!EnchantingCodeExporter methodsFor: 'initialization' stamp: 'CWB 8/19/2010 11:38'!
readTemplateFile: aFilename
	"Reads in the template file"

	| f |
	f _ FileStream readOnlyFileNamedOrNil: 
		(ScratchTranslator translationDir fullNameFor: aFilename).
	f ifNil: [
		DialogBoxMorph inform: 'File not found' withDetails: aFilename.
		^ nil].
	template _ f contentsOfEntireFile lines.
! !


!EnchantingCodeExporter methodsFor: 'accessing' stamp: 'CWB 7/28/2010 09:54'!
codeStream
	"Gets the stream that code is being stored on"

	^ codeStream! !

!EnchantingCodeExporter methodsFor: 'accessing' stamp: 'CWB 7/28/2010 09:55'!
codeStream: aStream
"Sets the stream to store code on.  If nil, creates a new code stream"

	aStream ifNil:
		[ codeStream _ WriteStream on: (String new: 5000) ]
	ifNotNil:
		[ codeStream _ aStream ]! !

!EnchantingCodeExporter methodsFor: 'accessing' stamp: 'CWB 7/21/2010 13:28'!
sprite
	"Grants access to the sprite being exported"

	^ sprite
! !

!EnchantingCodeExporter methodsFor: 'accessing' stamp: 'CWB 7/21/2010 13:28'!
sprite: aSprite
	"Sets the sprite to be operated on."

	sprite _ aSprite
! !


!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/19/2010 10:20'!
createUniqueFunctionNameFrom: baseName
	"Ensures that each function has a unique name"

	| appendix name |

	(functions includesKey: baseName) ifFalse: [^ baseName].

	"Keep trying the function name with a new number on the end"
	"until we come up with one that is not in use."

	appendix _ 2.

	[ true ] whileTrue:
	[
		name _ baseName, appendix asString.
		(functions includesKey: name) ifFalse: [^ name].
		appendix _ appendix + 1.
	] 
	! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/14/2010 22:28'!
exportArgument: aMorph
	"Takes the given argument to another block, translates if necessary, and outputs the result"

	argDepth _ argDepth + 1.

	"A number, a string, a dropdown -- anything except a code block will be an ArgMorph"
	(aMorph isKindOf: ArgMorph) ifTrue: [
		"codeStream nextPutAll: '[ArgMorph] '. " 
		aMorph exportArgWith: self.
	].

	"Any place a code tile has been inserted is a BlockMorph"
	(aMorph isKindOf: BlockMorph) ifTrue: [
		"codeStream nextPutAll: '[BlockMorph] '.  "
		"codeStream nextPut: $(."
		self exportBlock: aMorph.		"May need to flag that no indent applies"
		codeStream skip: -1.  "remove carriage return"
		"codeStream nextPut: $)."
	].

	argDepth _ argDepth - 1.

! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/18/2010 20:46'!
exportBlock: aBlock
	"Takes the given block, translates it to code, and outputs the result"

	| englishSpec codeSpec argMapping argIndex |

	self exportCommentAttachedTo: aBlock.

	aBlock isHatBlock 
		ifTrue: [ englishSpec _ aBlock englishCodeSpec ]
		ifFalse: [ englishSpec _ self englishSpecForSelector: aBlock selector. ].

	englishSpec ifNil: [ 
		codeStream nextPutAll: 
			'->NO ENGLISH FOR ' , 
			aBlock class asString asUppercase , 
			'<-'. 
		^ nil 
	].

	codeSpec _ self codeSpecFrom: englishSpec.
	argMapping _ aBlock class argPermutationForSpec: englishSpec withTranslation: codeSpec.

	self indent.

	argIndex _ 1.
	(CommandBlockMorph parseCommandSpec: codeSpec removeSpaces: false ) do: 
		[ :token |
			(CommandBlockMorph isArgSpec: token) 
				ifTrue: [ 
					| arg |
					"Look up the proper argument"
					arg _ aBlock argumentAt: (argMapping at: argIndex)	.
					self exportArgument: arg.
					argIndex _ argIndex + 1.
				] 
				ifFalse: [ codeStream nextPutAll: token. ].
		].

	codeStream cr.

! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/19/2010 11:54'!
exportCommentAttachedTo: aBlock
"Emits the comment attached to this block (if any)"

	| commentMorph commentText |

	'The inputLines are an array with a new line at every soft linebreak.'.
	'It also has hard line breaks wherever return was pressed.'.
	'outputLines are the same thing, without the line breaks' .

	commentMorph _ aBlock commentMorph.
	commentMorph ifNil: [ ^ nil ].

	commentText _ commentMorph commentMorph contents.

	self putComment: commentText on: codeStream maxLineLength: 60.


! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/19/2010 15:43'!
exportGreenFlagScript: aBlock
	"Emits code for a script triggered by startup"

	self exportCommentAttachedTo: aBlock.
	self exportScript: aBlock 
		usingName: 'WhenGreenFlagClicked'
		templateSection: 'BEGIN WHEN GREEN FLAG CLICKED TEMPLATE'.

	! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/19/2010 15:52'!
exportNestedBlock: aNestedBlock
	"Outputs a compound statement -- a series of statements that act rather like one statement."

	self outputTranslatedLineOf: 'nested block start'.
	indentationLevel _ indentationLevel + 1.

	aNestedBlock 
		ifNil: [
			self outputTranslatedLineOf: 'empty nested block'.
		]
		ifNotNil: [
			aNestedBlock exportCodeWith: self.
		].
 
	indentationLevel _ indentationLevel - 1.
	self outputTranslatedLineOf: 'nested block end'.
! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/21/2010 12:40'!
exportScript: aBlock usingName: baseName templateSection: aSection
	"Emits a function with a given base name."

	| functionName |

	functionName _ self createUniqueFunctionNameFrom: baseName.

	"Create a new code stream"
	self codeStream: nil.

	self exportCommentAttachedTo: aBlock.

	self processTemplateSection: aSection ontoStream: codeStream withCommands: 
	[ :command :prevToken |
		command = 'FUNCTIONNAME' ifTrue: [ codeStream nextPutAll: functionName ].
		command = 'FUNCTIONBODY' ifTrue: 
		[
			self usingIndent: prevToken 
				on: codeStream 
				execute: [ 
					aBlock nextBlock 
						ifNil: [
							self outputTranslatedLineOf: 'empty nested block'.
						]
						ifNotNil: [
							aBlock nextBlock exportCodeWith: self.
						].
				]
		].
	].
		
	"Store everything into a function structure"
	functions at: functionName put: 
		(EnchantingFunction 
			newWithSource: codeStream contents
			type: #greenFlag
			index: functionIndex).
	functionIndex _ functionIndex + 1.! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/19/2010 11:48'!
indent

	"Outputs indentation for the current line of code"

	self indentOn: codeStream.! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/19/2010 16:01'!
indentOn: aStream

	"Outputs indentation for the current line of code"

	| indentText |

	indentText _ self codeSpecFrom: 'indent'.

	argDepth == 0 ifTrue: 
	[ 
		aStream nextPutAll: additionalIndent.
		indentationLevel timesRepeat: [ aStream nextPutAll: indentText ].
	]! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/21/2010 12:04'!
printEachFunctionOn: aStream usingTemplateSection: aSectionName
	"The template can call for a section of code to be invoked for each function/thread.  This function makes it happen."

	functions keysDo: [:key |
		| func |

		func _ functions at: key.		

		self processTemplateSection: aSectionName ontoStream: aStream withCommands: 
		[:command :prevToken |
			command = 'FUNCTIONNAME' ifTrue: [ aStream nextPutAll: key ].
			command = 'FUNCTIONINDEX' ifTrue: [ aStream nextPutAll: func index asString ].
		].
	].

	


! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/26/2010 20:07'!
printOn: aStream
	"Prints the exported code onto the given stream"

	"Bail out if we don't have all the information we need"
	template ifNil: [aStream nextPutAll: 'Could not read template file.'].
	dict ifNil: [aStream nextPutAll: 'Could not read dictionary file.'].
	((template isNil) or: [dict isNil]) ifTrue: [^nil].

	additionalIndent _ ''.

	self processTemplateSection: 'BEGIN FILE TEMPLATE' ontoStream: aStream withCommands: 
	[:command :prevToken |
		command = 'ENCHANTING HEADER' ifTrue: [ self putHeaderOn: aStream ].
		command = 'IMPORTS' ifTrue: [ "to do" ].
		command = 'BASEFILENAME' ifTrue: [ aStream nextPutAll: outputFileBaseName ].
		command = 'FUNCTIONS' ifTrue: 
		[
			self usingIndent: prevToken on: aStream execute: [ self putFunctionsOn: aStream ].
		].
		('USE "*" FOR EACH FUNCTION' match: command) ifTrue:
		[
			| sectionName |
			sectionName _ (command findTokens: '"') second.	
			self usingIndent: prevToken 
				on: aStream
				execute: [ self printEachFunctionOn: aStream usingTemplateSection: sectionName. ] 
		]
	].
			

	


! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/19/2010 16:03'!
processScript: aScript

	"Takes a script (a hat block that starts a list of blocks) and processes it for output from the exporter"

	"The first thing the hat block will do is create a new function, complete with code stream"

	aScript exportCodeWith: self.

! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/19/2010 11:55'!
putComment: aComment on: aStream maxLineLength: maxLength

	"Takes a comment which may have multiple lines and puts it onto a stream."

	(aComment withNoLineLongerThan: maxLength) lines do: 
	[:line |
		self putSingleLineComment: line on: aStream
	].! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/19/2010 11:56'!
putFunctionsOn: aStream

	functions do: [:f | aStream nextPutAll: f source; cr. ]
! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/19/2010 12:26'!
putHeaderOn: aStream
	"Emits information about enchanting, and places it as a comment on the stream"

	self putComment: 'Exported from Enchanting' on: aStream maxLineLength: 60.! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/19/2010 11:49'!
putSingleLineComment: aComment on: aStream 
	"Outputs a comment onto a stream"

	self indentOn: aStream.
	aStream nextPutAll: (self translationOf: 'single-line comment start' ifAbsent: ['']).
	aStream nextPutAll: aComment.
	aStream nextPutAll: (self translationOf: 'single-line comment end' ifAbsent: ['']).
	aStream cr.! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/21/2010 12:36'!
usingIndent: aNewIndent on: aStream execute: aBlock 
	"Pushes a new indentation level, executes the code, pops the indentation level.  Useful for formatting replaced code within templates"

	| prevIndent |

	prevIndent _ additionalIndent.
	additionalIndent _ aNewIndent.
	
	"The indent will already be on the stream; remove it so it is not there twice"
	aStream skip: (aNewIndent size) * -1.
	"aStream nextPutAll: 'REMOVED ->', aNewIndent, '<- (', aNewIndent size asString, ')'."

	aBlock value.

	"remove the carriage return following the command to insert code from a template"
	aStream skip: -1.

	additionalIndent _ prevIndent.	

! !


!EnchantingCodeExporter methodsFor: 'templates' stamp: 'CWB 8/21/2010 12:31'!
processTemplateSection: aSectionName ontoStream: aStream withCommands: commandBlock
	"Goes through the template (a collection of lines), emits text or acts on meta commands."

	| sectionTemplate metaChar isCommand prevToken |

	metaChar _ self templateMetaCharacter.
	isCommand _ false.
	sectionTemplate _ self templateSection: aSectionName.
	prevToken _ ''.

	sectionTemplate do: [:line |
		| tokens |

		self indentOn: aStream.
		prevToken _ ''.

		tokens _ line findTokens: metaChar keep: metaChar.
		tokens do: [:token |
			token = metaChar 
				ifTrue: [ isCommand _ isCommand not ]
				ifFalse: [
					isCommand 
						ifTrue: [ commandBlock value: token value: prevToken ]
						ifFalse: [
							prevToken _ token.
							aStream nextPutAll: token.
						]
				] 
		].
	
		"Put a return on the line if the line was not just one command."
		"If it was one command, it will be exactly three tokens (meta, command, meta)."
		"No other situation will produce three tokens."
		(tokens size = 3) ifFalse: [aStream cr.]
	]
! !

!EnchantingCodeExporter methodsFor: 'templates' stamp: 'CWB 8/19/2010 15:11'!
templateMetaCharacter
	"Returns the template meta-character (as a string)."
	"(This character delimits portions of the template that will be filled in based on what the user has created in enchanting.)"

	^ self translationOf: 'template meta sequence' ifAbsent: ['@'].! !

!EnchantingCodeExporter methodsFor: 'templates' stamp: 'CWB 8/19/2010 15:15'!
templateSection: aTitle
	"Returns a portion of the template bounded by aTitle and 'END TEMPLATE', both within metacharacters"

	| metaChar startLine endLine startIndex endIndex |

	metaChar _ self templateMetaCharacter.
	startLine _ metaChar, aTitle, metaChar.
	endLine _ metaChar, 'END TEMPLATE', metaChar.

	startIndex _ template indexOf: startLine.
	startIndex = 0 ifTrue: [^nil].

	endIndex _ template indexOf: endLine startingAt: startIndex ifAbsent: [ 0 ].
	endIndex = 0 ifTrue: [^nil].

	^ template copyFrom: startIndex + 1 to: endIndex - 1. ! !


!EnchantingCodeExporter methodsFor: 'translating' stamp: 'CWB 8/18/2010 21:02'!
codeSpecFrom: english
	"Returns the specification for the block in the computer language we are exporting to"

	^ dict at: english ifAbsent: [ '>>?', english, '?<<' ].! !

!EnchantingCodeExporter methodsFor: 'translating' stamp: 'CWB 7/27/2010 19:36'!
englishSpecForSelector: aSelector
	"Returns the specification for the block in the native english"

	^ self sprite blockSpecForSelector: aSelector.	! !

!EnchantingCodeExporter methodsFor: 'translating' stamp: 'CWB 8/19/2010 11:07'!
importTranslation: filename
	"Reads a translation dictionary from the specified file"
 
	^ dict _ ScratchTranslator importTranslation: filename
	! !

!EnchantingCodeExporter methodsFor: 'translating' stamp: 'CWB 8/16/2010 21:16'!
outputTranslatedLineOf: aTerm
	"Outputs the translation of a given term (if it exists) as a whole line."

	(dict includesKey: aTerm)
		ifTrue: [
			self indent; outputTranslationOfTerm: aTerm.
			codeStream cr. 
		]! !

!EnchantingCodeExporter methodsFor: 'translating' stamp: 'CWB 8/14/2010 21:58'!
outputTranslationOfTerm: englishTerm
	"Takes an english term (such as appears in a choice argument morph), and translates it onto the code stream"

	codeStream nextPutAll: (self codeSpecFrom: englishTerm).! !

!EnchantingCodeExporter methodsFor: 'translating' stamp: 'CWB 8/18/2010 21:02'!
translationOf: english ifAbsent: block
	"Returns the specification for the block in the computer language we are exporting to"

	^ dict at: english ifAbsent: block! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

EnchantingCodeExporter class
	instanceVariableNames: ''!

!EnchantingCodeExporter class methodsFor: 'instance creation' stamp: 'CWB 8/19/2010 11:29'!
newWithSprite: aSprite exporter: baseFileName 
	"Reads in the given file which has instructions on how to translate"

	^ self new
		importTranslation: baseFileName, '.exporter';
		readTemplateFile: baseFileName, '.template';
		sprite: aSprite

! !

!EnchantingCodeExporter class methodsFor: 'instance creation' stamp: 'CWB 8/26/2010 20:00'!
newWithSprite: aSprite exporter: baseFileName outputAs: baseOutputName
	"Reads in the given file which has instructions on how to translate"

	^ self new
		importTranslation: baseFileName, '.exporter';
		readTemplateFile: baseFileName, '.template';
		outputFileBaseName: baseOutputName;
		sprite: aSprite

! !
